package de.tuberlin.cs.ni.neubauer.mpcd;

import java.io.IOException;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.PosixParser;

import de.tuberlin.cs.ni.neubauer.mpcd.assignments.CommunityAssignments;
import de.tuberlin.cs.ni.neubauer.mpcd.edges.EdgeContainerFactory;
import de.tuberlin.cs.ni.neubauer.mpcd.edges.ListEdgeContainerFactory;
import de.tuberlin.cs.ni.neubauer.mpcd.edges.MurataListEdgeContainerFactory;
import de.tuberlin.cs.ni.neubauer.mpcd.gain.DefaultGainTrackManager;
import de.tuberlin.cs.ni.neubauer.mpcd.gain.GainTrackManager;
import de.tuberlin.cs.ni.neubauer.mpcd.gain.GainTrackerFactory;
import de.tuberlin.cs.ni.neubauer.mpcd.gain.PythonGainTrackManager;
import de.tuberlin.cs.ni.neubauer.mpcd.io.ConsoleExporter;
import de.tuberlin.cs.ni.neubauer.mpcd.io.DataContainer;
import de.tuberlin.cs.ni.neubauer.mpcd.io.FileExporter;
import de.tuberlin.cs.ni.neubauer.mpcd.io.FileImporter;
import de.tuberlin.cs.ni.neubauer.mpcd.listeners.GainCheckListener;
import de.tuberlin.cs.ni.neubauer.mpcd.listeners.IntermediateOutputListener;
import de.tuberlin.cs.ni.neubauer.mpcd.listeners.IterationOutputListener;
import de.tuberlin.cs.ni.neubauer.mpcd.listeners.MergeOutputListener;
import de.tuberlin.cs.ni.neubauer.mpcd.listeners.ModularityCheckListener;
import de.tuberlin.cs.ni.neubauer.mpcd.listeners.UpdateCountListener;
import de.tuberlin.cs.ni.neubauer.mpcd.modularity.ModularityState;
import de.tuberlin.cs.ni.neubauer.mpcd.modularity.MurataModularityState;
//import de.tuberlin.cs.ni.neubauer.mpcd.modularity.OMMS2;
import de.tuberlin.cs.ni.neubauer.mpcd.modularity.OptimizedMurataModularityState;

/**
 * Multi-Partite Community Detection
 * main class
 * 
 * @author nn
  */
public class MPCD
{
	
	
	public static CommunityAssignments getResultOptimized(DataContainer data, CommandLine cmd) throws Exception
	{
		return getResult(getOptimizedMurataModularity(data, cmd, new MurataListEdgeContainerFactory()), cmd);
	}

	
	public static CommunityAssignments getResultBeforeOptimized(DataContainer data, CommandLine cmd) throws Exception
	{
		return getResult(getMurataModularity(data, cmd, new ListEdgeContainerFactory()), cmd);
	}

	
	public static CommunityAssignments getResult(DataContainer data, CommandLine cmd) throws Exception
	{
		return getResult(getMurataModularity(data, cmd, new MurataListEdgeContainerFactory()), cmd);
	}
	
	
	public static GainTrackerFactory getGainTrackerFactory(CommandLine cmd)
	{
		// if option "py", use GainTrackerManagers with a backwards-compatible tie-breaker method		
		GainTrackerFactory gainTrackerFactory = null;
		final Boolean verbose = cmd.hasOption("v");
		if (cmd.hasOption("py"))
		{
			gainTrackerFactory = new GainTrackerFactory()
			{
				public GainTrackManager getGainTrackManager(int numDimensions)
				{
					return new PythonGainTrackManager(numDimensions, verbose);
				}
			};
		}
		else
		{
			gainTrackerFactory = new GainTrackerFactory()
			{
				public GainTrackManager getGainTrackManager(int numDimensions)
				{
					return new DefaultGainTrackManager(numDimensions, verbose);
				}
			};
		}
		return gainTrackerFactory;
	}
	
	public static ModularityState getMurataModularity(DataContainer data, CommandLine cmd, EdgeContainerFactory edgeContainerFactory)
	{
		return new MurataModularityState(data, edgeContainerFactory, getGainTrackerFactory(cmd), !cmd.hasOption("weightDU"));
	}
	
	public static ModularityState getOptimizedMurataModularity(DataContainer data, CommandLine cmd, EdgeContainerFactory edgeContainerFactory)
	{
		return new OptimizedMurataModularityState(data, edgeContainerFactory, getGainTrackerFactory(cmd), !cmd.hasOption("weightDU"));
	}

	
	public static CommunityAssignments getResult(ModularityState ms, CommandLine cmd) throws Exception
	{
		System.out.println("initial modularity: " + ms.getCurrentModularity());
		Optimizer o = new Optimizer();
		
		
		// add listeners depending on command-line parameters
		if (cmd!=null)
		{
			if (cmd.hasOption("v"))
			{
				o.addListener(new IterationOutputListener());
				o.addListener(new UpdateCountListener(cmd.hasOption("oc")?cmd.getOptionValue("o")+".count.txt":"",
						ms.getNumDimensions()));
			}
			if (cmd.hasOption("c"))
			{
				o.addListener(new ModularityCheckListener());
				o.addListener(new GainCheckListener());		
			}
			if (cmd.hasOption("om"))
			{
				try
				{
					o.addListener(new MergeOutputListener(cmd.getOptionValue("o")+".joins.txt", ms.getData()));
				}
				catch (Exception e)
				{
					System.err.println("Exception initializing MergeOutputListener: " + e.getMessage());
				}
			}
			if (cmd.hasOption("oi"))
			{
				int interval = 1;
				if (cmd.hasOption("oii"))
				{
					try
					{
						interval = Integer.parseInt(cmd.getOptionValue("oii"));
					}
					catch (Exception e)
					{
						System.err.println(String.format("Error interpreting interval argument oii \"%s\" as integer, assuming interval 1", cmd.getOptionValue("oii")));
					}
				}
				String filename = "";
				if (cmd.hasOption("oif"))
				{
					filename = cmd.getOptionValue("oi");
				}
				else if (cmd.hasOption("o"))
				{
					filename = cmd.getOptionValue("o");
				}
				if (filename=="")
				{
					System.err.println("Please provide either a specific filename for intermediate community assignments via -oif, or a general output filename via -o");
				}
				else
				{
					o.addListener(new IntermediateOutputListener(filename, interval));
				}
			}
		
		}
		
		// main portion of work will be performed inside the next two calls
		if (o.optimize(ms))
		{
			if (cmd.hasOption('o'))
			{
				try
				{
					new IntermediateOutputListener(cmd.getOptionValue("o"), 0).onStep(o, null);
				}
				catch (Exception e)
				{
					throw (new RuntimeException(e));
				}
			}
			
			while (o.continueOptimize());
		}
		
		// final matters
		o.shutDown();
		System.out.println(String.format("best found modularity: %1.3f in step %d",
										o.getBestCommunityAssignments().getValue(),
										o.getBestCommunityAssignments().getStep()));
		return o.getBestCommunityAssignments();
	}
	
	public static Options getOptions()
	{
		Options options = new Options();
		
		// user options
		options.addOption("v", false, "verbose output");
		options.addOption("o", true, "output file (will be postfixed with .[dimension])");
		
		options.addOption("om", false, "output merges into file in format dim\\tsource\\ttarget\\tmodularity\\tstep");
		
		options.addOption("oi", false, "output intermediate community assignments");
		options.addOption("oif", false, "output file for intermediate community assignments");
		options.addOption("oii", true, "output interval for intermediate community assignments (num of steps between assignments)");
		
		options.addOption("oc", false, "output file for update counts");
		
		options.addOption("w", false, "input file contains weighted edges");
		options.addOption("weightDU", false, "weight edges between 0th and 1st dimension (document/user typically)");
		
		// debug options
		options.addOption("c", false, "check intermediate results (debug - slow)");
		options.addOption("tm", true, "mini test data (debug)");
		options.addOption("tf", true, "file test data (debug)");
		options.addOption("bo", false, "before optimization (debug)");
		options.addOption("py", false, "compatibility to previous python implementation (debug - slightly slower)");
		options.addOption("ce", false, "cutting edge (current development optimization method)");
		
		return options;
	}
	
	
	public static void doit(String arg) throws Exception
	{
		doit(arg.split(" "));
	}
	
	public static void doit(String[] args) throws Exception
	{
		Options options = getOptions();
		CommandLineParser parser = new PosixParser();
		CommandLine cmd = parser.parse(options, args);
		
		DataContainer data = null;
		
		// debug cases
		if (cmd.hasOption("tm"))
		{
			try
			{
				int index = Integer.parseInt(cmd.getOptionValue("tm"));
				data = MPCDExperimental.miniData(index);
			}
			catch (NumberFormatException e)
			{
				System.out.println(e.getMessage());
			}
		}
		else if (cmd.hasOption("tf"))
		{
			try
			{
				int index = Integer.parseInt(cmd.getOptionValue("tf"));
				data = MPCDExperimental.sampleFiles(index);
			}
			catch (Exception e)
			{
				System.out.println(e.getMessage());
			}
		}
		
		// normal usage: exactly one non-named argument - the input file
		else if (cmd.getArgList().size()==1)
		{
			try
			{
				data = new FileImporter((String)cmd.getArgList().get(0)).getEdges(cmd);
			}
			catch (IOException e)
			{
				System.out.println(e.getMessage());
			}
		}
		
		// catch bad parameterization
		if (data==null)
		{
			HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("java mpcd.jar FILE", options);
			return;
		}
		
		// do work
		CommunityAssignments assignments = null;
		
		if (cmd.hasOption("bo"))
			assignments = getResultBeforeOptimized(data, cmd);
		else if (cmd.hasOption("ce"))
			assignments = getResultOptimized(data, cmd);
		else
			assignments = getResult(data, cmd); 
		
		// output results
		if (cmd.hasOption("o"))
		{
			new FileExporter(cmd.getOptionValue("o")).writeResults(assignments, data);
		}
		else
		{
			new ConsoleExporter().writeResults(assignments, data);
		}
		
	}
	
	

	/**
	 * @param args
	 */
	public static void main(String[] args) throws Exception
	{
		doit(args);
		//String args_dings = "-o f:\files\workspace\python\tagging\output\bibsonomy_nospam\tag\41385\randomwalk\50_73295_c1\edges.txt -weightDU   -v  -c   -ce -oc f:\files\count.txt -om f:\files\compare_java.txt f:\files\workspace\python\tagging\output\bibsonomy_nospam\tag\41385\randomwalk\50_73295_c1\edges.txt";
		
		//doit(" -tm 3");
		
		
		//String args_synth = "-weightDU -v -c -tf 0 ";
		//doit(args_synth + " -o f:\\files\\edges -oc f:\\files\\count.txt -ce -om f:\\files\\merges.txt -omm -oi -oii 10");
		//doit(args_synth + " -oc f:\\files\\xcount.txt ");
		
	}	
}
