package de.tuberlin.cs.ni.neubauer.mpcd;

import java.util.ArrayList;
import java.util.List;

import de.tuberlin.cs.ni.neubauer.mpcd.assignments.CommunityAssignments;
import de.tuberlin.cs.ni.neubauer.mpcd.gain.GainTrackManager;
import de.tuberlin.cs.ni.neubauer.mpcd.listeners.ListenerException;
import de.tuberlin.cs.ni.neubauer.mpcd.listeners.OptimizationStepListener;
import de.tuberlin.cs.ni.neubauer.mpcd.modularity.ModularityState;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.JoinPair;

/**
 * takes a modularity state and finds the best community assignment 
 * @author nn
 *
 */
public class Optimizer
{
	List<OptimizationStepListener> listeners;
	public CommunityAssignments bestCommunities;
	public GainTrackManager gtm;
	public ModularityState modularity;
	public int step;
	
	public Optimizer()
	{
		this.listeners = new ArrayList<OptimizationStepListener>();
	}
	
	/**
	 * adds a listener to be called after each optimization step
	 * @param listener
	 */
	public void addListener(OptimizationStepListener listener)
	{
		this.listeners.add(listener);
	}
	
	/**
	 * sets up the optimization
	 * @param importer
	 * @return
	 */
	public Boolean optimize(ModularityState modularity)
	{
		this.modularity = modularity;
		this.gtm = this.modularity.getGainTrackManager();
		this.step = 0;
		return true;
	}
	
	CommunityAssignments getAssignmentCopy()
	{
		return this.modularity.getCommunityAssignments().copyAssignments(this.modularity.getCurrentModularity(), this.step);
	}
	
	/**
	 * main optimization loop
	 * @return whether it can be called again (ie there is still stuff to join and we're exiting 
	 *         just because of a listener veto)
	 */
	public Boolean continueOptimize()
	{
		double lastGain = -1;
		JoinPair next = gtm.nextJoin();
		this.bestCommunities = getAssignmentCopy();
		while (next != null)
		{			
			// if we're about to get worse, save current state (if we got worse in the previous step, 
			// no need to save, keep the current one)			
			if (next.gain<0 && lastGain>0)
				if (this.bestCommunities == null || this.modularity.getCurrentModularity()>this.bestCommunities.getValue())
				{
					this.bestCommunities = getAssignmentCopy();
				}
			
			lastGain = next.gain;
			this.step += 1;			
			
			
			// calling the main worker routines
			this.modularity.join(next, gtm);
			
			// listeners (can request abort)
			Boolean abort = false;			
			for (OptimizationStepListener listener: this.listeners)
			{
				try
				{		
					abort = abort || listener.onStep(this, next);
				}
				catch (ListenerException e)
				{
					throw (new RuntimeException(e));
				}
				catch (Exception e)
				{
					System.err.println(String.format("Error in listener %s: %s", listener, e.getMessage()));
				}
			}
			if (abort)
				break;
			
			// ok, next please
			next = gtm.nextJoin();
		}

		// ok let's see if we need to make another copy of the state
		// this is the case if we leave the loop without the last join being the first negative-gain join 
		double modValueNow = this.modularity.getCurrentModularity();
		if (this.bestCommunities==null)
			this.bestCommunities = getAssignmentCopy();
		else if (modValueNow>this.bestCommunities.getValue())
			this.bestCommunities = getAssignmentCopy();
		return false;
	}
	
	public void shutDown()
	{
		for (OptimizationStepListener listener: this.listeners)
		{
			listener.shutDown();
		}
	}
	
	/**
	 * 
	 * @return the best community assignment obtained (so far)
	 */
	public CommunityAssignments getBestCommunityAssignments()
	{
		return this.bestCommunities;
	}
	
	
}
