package de.tuberlin.cs.ni.neubauer.mpcd.assignments;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.Pair;


/**
 * represents a mapping from elements to partitions
 * @author nn
 *
 */
public class CommunityAssignment
{
	Map<Integer,List<Integer>> partitions;
	Map<Integer,Integer> sizes;

	/**
	 * no initialization
	 */
	public CommunityAssignment()
	{
		this.partitions = new HashMap<Integer, List<Integer>>();
		this.sizes = new HashMap<Integer, Integer>();			
	}
	
	/**
	 * initialize from list of nodes
	 * @param nodes
	 */
	public CommunityAssignment(Collection<Integer> nodes)
	{
		this();
		initializeFromNodes(nodes);
	}
	
	/**
	 * create a deep copy of another CommunityAssignment
	 * @param original
	 */
	public CommunityAssignment(CommunityAssignment original)
	{
		this();
		for (Map.Entry<Integer, List<Integer>> entry: original.partitions.entrySet())
			this.partitions.put(entry.getKey(), entry.getValue());
		for (Map.Entry<Integer, Integer> entry: original.sizes.entrySet())
			this.sizes.put(entry.getKey(), entry.getValue());
	}
	
	
	/**
	 * initialize from previous mapping 
	 * if partial grouping of nodes is already known before algorithm starts, e.g. by simple clustering algorithm
	 * @param partitions
	 */
	public CommunityAssignment(Map<Integer,List<Integer>> partitions)
	{
		this.partitions = partitions;
		this.sizes = new HashMap<Integer, Integer>();
		for (int node: this.partitions.keySet())
			this.sizes.put(node, this.partitions.get(node).size());
	}
	
	public void initializeFromNodes(Collection<Integer> nodes)
	{
		for (int node: nodes)
		{
			ArrayList<Integer> al = new ArrayList<Integer>();
			al.add(node);
			this.partitions.put(node, al);
			this.sizes.put(node, 1);
		}
	}
	
	/**
	 * merge two communities
	 * 
	 * @param old the community that doesn't exist anymore after the merger
	 * @param new the community that the nodes are added to
	 */
	public void join(Pair pair)
	{
		int oldCommunity = pair.bigger_element;
		int newCommunity = pair.smaller_element;
		if (!this.partitions.containsKey(oldCommunity))
			throw (new RuntimeException("oldCommunity " + oldCommunity + " not in Partitioning"));
		if (!this.partitions.containsKey(newCommunity))
			throw (new RuntimeException("newCommunity " + newCommunity + " not in Partitioning"));
		this.partitions.get(newCommunity).addAll(this.partitions.get(oldCommunity));
		this.partitions.remove(oldCommunity);
		this.sizes.put(newCommunity, this.sizes.get(newCommunity)+this.sizes.get(oldCommunity));
		this.sizes.remove(oldCommunity);
	}
	
	/**
	 * get the list of nodes associated with a community 
	 * 
	 * @param community
	 * @return
	 */
	public List<Integer> getNodes(int community)
	{
		// TODO should this be safe from manipulation?
		if (!this.partitions.containsKey(community))
			throw (new RuntimeException("community " + community + " not in Partitioning"));
		return this.partitions.get(community);
	}

	/**
	 * returns the number of nodes in a community
	 * @param community
	 * @return
	 */
	public int getSize(int community)
	{
		if (!this.partitions.containsKey(community))
			throw (new RuntimeException("community " + community + " not in Partitioning"));
		return this.sizes.get(community);
	}
	

	/**
	 * 
	 * @return a map from nodes to community IDs
	 */
	public Map<Integer, Integer> byNode()
	{
		HashMap<Integer, Integer> nodeMap = new HashMap<Integer, Integer>();
		for (int community: this.partitions.keySet())
			for (int node: this.partitions.get(community))
				nodeMap.put(node, community);
		return nodeMap;
	}
	
	/**
	 * 
	 * @return a map from community IDs to associated nodes
	 */
	public Map<Integer, List<Integer>>byCommunity()
	{
		return this.partitions;
	}
	
	public CommunityAssignment deepCopy()
	{
		return new CommunityAssignment(this);
	}
}
