package de.tuberlin.cs.ni.neubauer.mpcd.edges;

import java.util.Collection;
import java.util.Map;
import java.util.Set;

import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.CommunityWithStrength;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.DoublePair;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.Pair;

/**
 * storing edges from communities of one domain to communities of another domain
 * 
 * @author nn
 *
 */
public interface EdgeContainer
{
	/** add an edge from <i>from</i> to <i>to</i> 
	 * 
	 * @param from
	 * @param to
	 */
	public void add(int from, int to);
	
	/**
	 * specify connection strength
	 * @param from
	 * @param to
	 * @param strength
	 */
	public void add(int from, int to, double strength);
	
	/**
	 * merge two elements from the "from" domain
	 * @param pair
	 */
	public void merge_from(Pair pair);
	
	/**
	 * merge two elements from the "to" domain
	 * @param pair
	 */
	public void merge_to(Pair pair);
	
	/**
	 * the total number of edges from community
	 * @param community
	 * @return
	 */
	public double getSize(int community);
	
	/**
	 * normalization coefficient (1/(2*sum of all edge weights))
	 * @return
	 */
	public double norm();
	
	/**
	 * 
	 * @return all "from" communities (keys of HashMap)
	 */
	public Collection<Integer> allFromCommunities();
	
	/**
	 * 
	 * @return
	 */
	public Set<Integer> allToCommunities(int from);
	
	/**
	 * returns a pair indicating the 
	 * @param communityFrom
	 * @param pairTo
	 * @return
	 */
	public DoublePair connectionsTo(int communityFrom, Pair pairTo);
	
	
	public double getStrength(int from, int to);
		
		
	// below: methods required for Murata (single best corresponding community)
	
	/**
	 * finds for each "from" community the strongest associated to community 
	 * @return fromCommunity -> toCommunity (with strength)
	 */
	public Map<Integer, CommunityWithStrength> getBestCommunities(EdgeContainer toContainer);
	
	
	/**
	 * return the best matching to-community for a potentially joined community of element_1 and element_2
	 * i.e. the "from" dimension is the one being joint and we're looking for a new corresponding community
	 * for the new, joint community 
	 * 
	 * must consider joint connections from the two joinFrom-communities as a single community 
	 * 
	 * @param joinFrom
	 * @param toContainer
	 * @return
	 */
	public CommunityWithStrength getBestOtherCommunity(Pair joinFrom,
													   EdgeContainer toContainer,
													   Map<Integer, CommunityWithStrength> corresponding);
	
	/**
	 * get the best community from the dimension that's being joint
	 * i.e., the "to" dimension is the one being joint and we're trying to see if it gains or loses
	 * correspondence status for another (the, here, "from") dimension
	 * 
	 * it's called OneEmpty because it's only called when one of the joint communities has 
	 * no connections to fromElement so we don't have to consider edge additions like in bestOtherCommunity
	 * 
	 * @param fromElement
	 * @param joinTo
	 * @param toContainer EdgeContainer from "to" dimension to "from" - for weights
	 * @param jointSize overrides weight values in toContainer for to-be-joint community
	 * @return
	 */
	public CommunityWithStrength getBestJointCommunityOneEmpty(int fromElement,
													   		   Pair joinTo,
													   		   EdgeContainer toContainer,
													   		   double jointSize);


	/**
	 * 
	 * @param myCommunities
	 * @param ignoreOtherCommunities
	 * @return
	 */
	public Set<Integer> getConnectedFromAny(Set<Integer> myCommunities);
}
