package de.tuberlin.cs.ni.neubauer.mpcd.edges;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.CommunityWithStrength;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.DoublePair;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.Pair;

public class ListEdgeContainer implements EdgeContainer
{
	HashMap<Integer, Map<Integer, Double>> edges;
	Boolean unweighted = false;
	double norm=-1;
	double[] sizes;
	
	public ListEdgeContainer()
	{
		this.edges = new HashMap<Integer, Map<Integer, Double>>();
	}
	
	public ListEdgeContainer(Boolean unweighted)
	{
		this();
		this.unweighted = unweighted;
	}
	
	@Override
	public void add(int from, int to)
	{
		add(from, to, 1);
	}

	@Override
	public void add(int from, int to, double strength)
	{
		// cached size variables dirty
		this.norm = -1;
		
		Map<Integer, Double> f;
		if (!this.edges.containsKey(from))
		{
			f = new HashMap<Integer, Double>();
			this.edges.put(from, f);
		}
		else
		{
			f = this.edges.get(from);
		}
		if (unweighted || (!f.containsKey(to)))
			f.put(to, strength);
		else
			f.put(to, f.get(to)+strength);
	}	

	@Override
	public void merge_from(Pair pair)
	{
		int oldCommunity = pair.bigger_element;
		int newCommunity = pair.smaller_element;
		if (!this.edges.containsKey(oldCommunity))
			return;
		if (!this.edges.containsKey(newCommunity))
		{
			this.edges.put(newCommunity, this.edges.get(oldCommunity));
			this.edges.remove(oldCommunity);
			this.sizes[newCommunity] = this.sizes[oldCommunity];
			return;
		}
		Map<Integer, Double> newCommunityEdges = this.edges.get(newCommunity);
		for (Map.Entry<Integer, Double> e: this.edges.get(oldCommunity).entrySet())
		{
			int communityTo = e.getKey();
			double strength = e.getValue();
			if (newCommunityEdges.containsKey(communityTo))
				strength += newCommunityEdges.get(communityTo);
			newCommunityEdges.put(communityTo, strength);
		}
		this.edges.remove(oldCommunity);
		this.sizes[newCommunity] += this.sizes[oldCommunity];
		
	}
		

	@Override
	public void merge_to(Pair pair)
	{
		int oldCommunity = pair.bigger_element;
		int newCommunity = pair.smaller_element;
		
		for (Map.Entry<Integer, Map<Integer, Double>> e: this.edges.entrySet())
		{
			Map<Integer, Double> m = e.getValue();
			if (!m.containsKey(oldCommunity))
				continue;
			double strength = 0;
			
			if (m.containsKey(newCommunity))
				strength = m.get(newCommunity);
			m.put(newCommunity, strength + m.get(oldCommunity));
			m.remove(oldCommunity);
		}
		
	}

	@Override
	public double getSize(int community)
	{
		if (this.sizes==null)
		{
			int n = Collections.max(this.edges.keySet())+1;
			this.sizes = new double[n];
		}
		double size = sizes[community];
		if (size>0)
			return size;
		for (double strength: this.edges.get(community).values())
			size += strength;
		sizes[community] = size;
		return size;
	}

	@Override
	public double norm()
	{
		if (this.norm>-1)
			return this.norm;
		double s = 0;
		for (Map<Integer, Double> map: this.edges.values())
			for (double strength: map.values())
				s += strength;
		this.norm = 0.5/s;
		return this.norm;
	}

	@Override
	public Collection<Integer> allFromCommunities()
	{
		return this.edges.keySet();
	}
	

	@Override
	public Map<Integer, CommunityWithStrength> getBestCommunities(EdgeContainer toContainer)
	{
		Map<Integer, CommunityWithStrength> map = new HashMap<Integer, CommunityWithStrength>();
		
		for (Map.Entry<Integer, Map<Integer, Double>> entry: this.edges.entrySet())
			map.put(entry.getKey(), getBest(entry.getValue(), toContainer));
		
		return map;
	}
	
	
	@Override
	/**
	 * eigentlich kann nur folgendes passieren:
	 *   corrComm von smaller oder von bigger bleibt corrComm
	 *   eine Community, die mit beiden verbunden ist, wird neue CorrComm
	 *   
	 * d.h. communities, die weder corresponding noch in smaller UND bigger vertreten sind, muessen
	 * gar nicht erst betrachtet werden (koennte bei (!) beachtet werden )
	 */
	public CommunityWithStrength getBestOtherCommunity(Pair joinFrom,
			EdgeContainer toContainer,  
			Map<Integer, CommunityWithStrength> ignored)
	{
		Map<Integer, Double> overrideMap = new HashMap<Integer, Double>();
		Map<Integer, Double> connectionsFromSmaller = this.edges.get(joinFrom.smaller_element); 
		Map<Integer, Double> connectionsFromBigger = this.edges.get(joinFrom.bigger_element); 
		for (Map.Entry<Integer, Double> e: connectionsFromBigger.entrySet())
		{
			int toCommunity = e.getKey();
			if (connectionsFromSmaller.containsKey(toCommunity))
				overrideMap.put(toCommunity, e.getValue() + connectionsFromSmaller.get(toCommunity));
			else
				overrideMap.put(toCommunity, e.getValue());			/* (1) */
		}
		
		return getBest(this.edges.get(joinFrom.smaller_element), toContainer, overrideMap, null);
	}	
	

	@Override
	/**
	 * ok der Fall ist hier dass in joinTo zwei Communities stehen, die gemergt werden
	 * eine von denen ist bislang corresponding
	 * die andere hat keine verbindung
	 * die frage ist, ob die zielcommunity corresponding bleibt 
	 * die einzige mglichkeit ist, dass es bislang eine gleich verbundene Community gibt,
	 * die jetzt vorgezogen wird
	 * 
	 */
	public CommunityWithStrength getBestJointCommunityOneEmpty(int fromElement,
			Pair joinTo, EdgeContainer toContainer, double jointEdges)
	{

		Map<Integer, Double> overrideEdges = new HashMap<Integer, Double>();
		overrideEdges.put(joinTo.smaller_element, jointEdges);
		overrideEdges.put(joinTo.bigger_element, 0.0);
		Map<Integer, Double> overrideSize = new HashMap<Integer, Double>();
		overrideSize.put(joinTo.smaller_element, toContainer.getSize(joinTo.smaller_element) + 
												 toContainer.getSize(joinTo.bigger_element));
		CommunityWithStrength result = getBest(this.edges.get(fromElement), toContainer, overrideEdges, overrideSize);
		return result;
	}
	

	
			
	
	/**
	 * glorified argmax 
	 * 
	 * @param toMap
	 * @param toContainer
	 * @param overrideMap override values in toMap
	 * @param overrideSize override values in toContainer.getSize()
	 * @return
	 */
	CommunityWithStrength getBest(Map<Integer, Double> toMap, 	
								  EdgeContainer toContainer,
							      Map<Integer, Double> overrideMap,
							      Map<Integer, Double> overrideSize)
	{
		int bestCommunity = -1;
		double bestStrength = -1;
		double bestSize = -1;
		
		if (toMap.size()==0)
			throw (new RuntimeException("getBest called with empty map!"));
		
		Set<Integer> allCommunities = null;
		if (overrideMap==null)
			allCommunities = toMap.keySet();
		else
		{
			allCommunities = new HashSet<Integer>();
			allCommunities.addAll(toMap.keySet());
			allCommunities.addAll(overrideMap.keySet());
		}
		
		for (int community: allCommunities)
		{
			double strength = 0;
			if (overrideMap==null || !overrideMap.containsKey(community))
				strength = toMap.get(community);
			else
				strength = overrideMap.get(community);
							
			if (strength<bestStrength)
				continue;
			
			
			double newSize = 0;
			if (overrideSize==null || !overrideSize.containsKey(community))
				newSize = toContainer.getSize(community);
			else
				newSize = overrideSize.get(community);
			
			if (strength>bestStrength)
			{
				bestStrength = strength;
				bestCommunity = community;
				bestSize = newSize;
				continue;
			}
			
			
			// if we are here, current and best community are of equal strength
			
			// decide for smaller community
			
			if (newSize>bestSize)
				continue;
			
			if (newSize<bestSize)
			{
				bestStrength = strength;
				bestCommunity = community;
				bestSize = newSize;			
			}
			
			
			// if we are here, current and best community are also of equal strength
			
			// choose smaller index just to make things deterministic as opposed to dependent
			// on order in Map traversal
			if (bestCommunity>community)
			{
				bestStrength = strength;
				bestCommunity = community;
				bestSize = newSize;			
			}
			
		}
		
		return new CommunityWithStrength(bestCommunity, bestStrength);
	}
	

	/**
	 * convenience method
	 */
	CommunityWithStrength getBest(Map<Integer, Double> toMap, EdgeContainer toContainer)
	{
		return getBest(toMap, toContainer, null, null);
	}


	
	@Override
	public DoublePair connectionsTo(int communityFrom, Pair pairTo)
	{
		Map<Integer, Double> connectionsFrom = this.edges.get(communityFrom);
		DoublePair result = new DoublePair();
		if (connectionsFrom.containsKey(pairTo.smaller_element))
			result.smaller_element = connectionsFrom.get(pairTo.smaller_element);
		if (connectionsFrom.containsKey(pairTo.bigger_element))
			result.bigger_element = connectionsFrom.get(pairTo.bigger_element);		
		return result;
	}

	@Override
	public Set<Integer> allToCommunities(int from)
	{
		return this.edges.get(from).keySet();
	}
	
	@Override
	public String toString()
	{
		StringBuilder sb = new StringBuilder();
		
		for (Map.Entry<Integer, Map<Integer, Double>> entry: this.edges.entrySet())
		{
			sb.append(String.format("%d: %s,", entry.getKey(), entry.getValue()));
		}
		
		return sb.toString();
	}

	@Override
	public Set<Integer> getConnectedFromAny(Set<Integer> myCommunities) 
	{
		Set<Integer> result = new HashSet<Integer>();
		for (int myCommunity: myCommunities)
			result.addAll(allToCommunities(myCommunity));
		
		return result;
	}

	@Override
	public double getStrength(int from, int to) 
	{
		if (!this.edges.containsKey(from))
			return 0;
		Map<Integer, Double> fromMap = this.edges.get(from);
		if (!fromMap.containsKey(to))
			return 0;
		return fromMap.get(to);
	}

		
}
