package de.tuberlin.cs.ni.neubauer.mpcd.edges;

import java.util.List;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;


import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.CommunityWithStrength;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.Pair;

/**
 * a special ListEdgeContainer containing a list of best Communities to speed up getBestJointCommunityOneEmpty
 * 
 * with a 10/10/1000-HG, getBestJointCommunityOneEmpty gets called about a million times
 * and the difference is worth 6 seconds which is about a quarter of the overall runtime of the original version
 * 
 * saves about a third in many standard cases
 * 
 * 20_10_2000_0.950: 1626 - 200 - 170 s (python, unoptimized, this one)
 * 
 * @author nn
 *
 */
public class MurataListEdgeContainer extends ListEdgeContainer 
{

	Map<Integer, List<Integer>> bestCommunities;
	
	public MurataListEdgeContainer()
	{
		super();
		this.bestCommunities = new HashMap<Integer, List<Integer>>();
	}
	
	public MurataListEdgeContainer(Boolean unweighted)
	{
		super(unweighted);
		this.bestCommunities = new HashMap<Integer, List<Integer>>();
	}
	


	@Override
	public void merge_from(Pair pair)
	{
		int oldCommunity = pair.bigger_element;
		int newCommunity = pair.smaller_element;
		if (!this.edges.containsKey(oldCommunity))
			return;
		if (!this.edges.containsKey(newCommunity))
		{
			this.edges.put(newCommunity, this.edges.get(oldCommunity));
			this.edges.remove(oldCommunity);
			this.sizes[newCommunity] = this.sizes[oldCommunity];
			this.bestCommunities.put(newCommunity, this.bestCommunities.get(oldCommunity));
			this.bestCommunities.remove(oldCommunity);
			return;
		}
		Map<Integer, Double> newCommunityEdges = this.edges.get(newCommunity);
		for (Map.Entry<Integer, Double> e: this.edges.get(oldCommunity).entrySet())
		{
			int communityTo = e.getKey();
			double strength = e.getValue();
			if (newCommunityEdges.containsKey(communityTo))
				strength += newCommunityEdges.get(communityTo);
			newCommunityEdges.put(communityTo, strength);
		}
		this.edges.remove(oldCommunity);
		this.sizes[newCommunity] += this.sizes[oldCommunity];
		
		// recompute when needed
		this.bestCommunities.remove(oldCommunity);
		this.bestCommunities.remove(newCommunity);
	}
		

	@Override
	public void merge_to(Pair pair)
	{
		int oldCommunity = pair.bigger_element;
		int newCommunity = pair.smaller_element;
		
		for (Map.Entry<Integer, Map<Integer, Double>> e: this.edges.entrySet())
		{
			Map<Integer, Double> m = e.getValue();
			int fromCommunity = e.getKey();
			if (!m.containsKey(oldCommunity))
				continue;
			double strength = 0;
			
			this.bestCommunities.remove(fromCommunity); // this may be a rather conservative approach to caching, but it's save 
			if (m.containsKey(newCommunity))
				strength = m.get(newCommunity);
			m.put(newCommunity, strength + m.get(oldCommunity));
			m.remove(oldCommunity);
		}
		
	}

	

	@Override
	/**
	 * ok der Fall ist hier dass in joinTo zwei Communities stehen, die gemergt werden
	 * eine von denen ist bislang corresponding
	 * die andere hat keine verbindung
	 * die frage ist, ob die zielcommunity corresponding bleibt 
	 * die einzige mglichkeit ist, dass es bislang eine gleich verbundene Community gibt,
	 * die jetzt vorgezogen wird
	 * eine liste aller maximal verbundenen communities knnte also massiv helfen
	 * 
	 */
	public CommunityWithStrength getBestJointCommunityOneEmpty(int fromElement,
			Pair joinTo, EdgeContainer toContainer, double jointEdges)
	{
		List<Integer> best = null;
		
		if (!this.bestCommunities.containsKey(fromElement))
			best = computeBestCommunities(fromElement);			
		else
			best = this.bestCommunities.get(fromElement);
		
		if (best.size()==0)
			throw (new RuntimeException("best.size()==0!"));
		if (best.size()==1)
		{
			int oldCorrComm = best.get(0); 
			if (!joinTo.contains(oldCorrComm))
				throw (new RuntimeException(String.format("best contains one but not the former corresponding community..? %s, %d",
						joinTo,
						oldCorrComm)));
			// easiest case: corresponding community is the only one with sufficient links, can get as large as it likes
			return new CommunityWithStrength(oldCorrComm, this.edges.get(fromElement).get(oldCorrComm));
		}
		// find smallest of equivalently linked
		
		double smallestSize = -1;
		int smallestCommunity = 0;
		for(int community: best)
		{
			double size = toContainer.getSize(community);
			if (community==joinTo.bigger_element)
				size += toContainer.getSize(joinTo.smaller_element);
			else if (community==joinTo.smaller_element)
				size += toContainer.getSize(joinTo.bigger_element);
			if (smallestSize==-1 || smallestSize>size)
			{
				smallestSize = size;
				smallestCommunity = community;
			}
			else if (smallestSize==size)
			{
				if (smallestCommunity>community)
					smallestCommunity = community; // if in doubt take smaller ID
			}
		}
		
		
		return new CommunityWithStrength(smallestCommunity, this.edges.get(fromElement).get(smallestCommunity));
				
	}
	
	public List<Integer> computeBestCommunities(int fromElement)
	{
		List<Integer> best = new ArrayList<Integer>();
		double bestStrength = -1;
		
		Map<Integer, Double> myEdges = this.edges.get(fromElement);
		for (Map.Entry<Integer, Double> e: myEdges.entrySet())
		{
			int toCommunity = e.getKey();
			double strength = e.getValue();
			if (bestStrength>strength)
				continue;
			if (bestStrength<strength)
			{
				best = new ArrayList<Integer>();
				bestStrength = strength;
			}
			best.add(toCommunity);
		}
		// only start saving this if there is a community with connection > 1
		// otherwise we're just replicating a lot of data in memory during
		// initializtion of unweighted graphs
		// since connection strength only increases during optimization, we 
		// don't have to care about deleting this ever
		if (bestStrength>1)
			this.bestCommunities.put(fromElement, best);
		return best;
	}
	
	
	/**
	 * this is a bit tedious, but does the job pretty efficiently. 
	 * @param joinFrom
	 * @param toContainer
	 * @param connectionsFromSmaller
	 * @param connectionsFromBigger
	 * @param corr_smaller
	 * @param corr_bigger
	 * @return
	 */
	CommunityWithStrength newCurrentCorrComm(Pair joinFrom,
			EdgeContainer toContainer,
			Map<Integer, Double> connectionsFromSmaller,
			Map<Integer, Double> connectionsFromBigger,
			CommunityWithStrength corr_smaller,
			CommunityWithStrength corr_bigger)
	{			
		
		double newSmallerCorrStrength = corr_smaller.strength;
		if (connectionsFromBigger.containsKey(corr_smaller.community))
			newSmallerCorrStrength += connectionsFromBigger.get(corr_smaller.community);
		if (connectionsFromSmaller.get(corr_smaller.community)!=
			corr_smaller.strength)
			throw (new RuntimeException("aua"));
		
		double newBiggerCorrStrength = corr_bigger.strength;
		if (connectionsFromSmaller.containsKey(corr_bigger.community))
			newBiggerCorrStrength += connectionsFromSmaller.get(corr_bigger.community);
		
		
		// find out which one of the previous corrcoms remains if we only have to decide between the two
		int newCorr = -1;
		double newStrength = -1;
		if (newSmallerCorrStrength>newBiggerCorrStrength)
		{
			newCorr = corr_smaller.community;
			newStrength = newSmallerCorrStrength;
		}
		else if (newSmallerCorrStrength<newBiggerCorrStrength)
		{
			newCorr = corr_bigger.community;
			newStrength = newBiggerCorrStrength;
		}
		else 
		{
			double size_smaller = toContainer.getSize(corr_smaller.community);
			double size_bigger = toContainer.getSize(corr_bigger.community);
			if (size_smaller<size_bigger)
			{
				newCorr = corr_smaller.community;
				newStrength = newSmallerCorrStrength;	
			}
			else if (size_bigger<size_smaller)
			{
				newCorr = corr_bigger.community;
				newStrength = newBiggerCorrStrength;					
			}
			else if (corr_bigger.community<corr_smaller.community)
			{
				newCorr = corr_bigger.community;
				newStrength = newBiggerCorrStrength;			
			}
			else
			{
				newCorr = corr_smaller.community;
				newStrength = newSmallerCorrStrength;			
			}
		}
		return new CommunityWithStrength(newCorr, newStrength);
	}
	
	@Override
	/**
	 * eigentlich kann nur folgendes passieren:
	 *   corrComm von smaller oder von bigger bleibt corrComm
	 *   eine Community, die mit beiden verbunden ist, wird neue CorrComm
	 *   
	 * d.h. communities, die weder corresponding noch in smaller UND bigger vertreten sind, muessen
	 * gar nicht erst betrachtet werden (koennte bei (!) beachtet werden )
	 */
	public CommunityWithStrength getBestOtherCommunity(Pair joinFrom,
			EdgeContainer toContainer,
			Map<Integer, CommunityWithStrength> corresponding)
	{

		Map<Integer, Double> connectionsFromSmaller = this.edges.get(joinFrom.smaller_element); 
		Map<Integer, Double> connectionsFromBigger = this.edges.get(joinFrom.bigger_element); 
		CommunityWithStrength corr_smaller = corresponding.get(joinFrom.smaller_element);
		CommunityWithStrength corr_bigger= corresponding.get(joinFrom.bigger_element);

		// so baseline is the better of the two previous corresponding communities
		CommunityWithStrength newCurr = 
			newCurrentCorrComm(joinFrom, toContainer, connectionsFromSmaller, connectionsFromBigger, corr_smaller, corr_bigger);
		// TODO revert to int/double for fast processing (?)
		// TODO understand previous item ;)
		double bestSize = toContainer.getSize(newCurr.community);
		
		for (Map.Entry<Integer, Double> e: connectionsFromBigger.entrySet())
		{
			int toCommunity = e.getKey();			
			if (corr_smaller.community==toCommunity)
				continue; // already checked that
			if (corr_bigger.community==toCommunity)
				continue; // ditto
			
			if (!connectionsFromSmaller.containsKey(toCommunity))
				continue;
			double newStrength = e.getValue() + connectionsFromSmaller.get(toCommunity);
			if (newStrength<newCurr.strength)
				continue;
			
			// strength decision
			double newSize = toContainer.getSize(toCommunity); 
			if (newStrength>newCurr.strength)
			{
				newCurr = new CommunityWithStrength(toCommunity, newStrength);
				bestSize = newSize;
				continue;
			}
			
			// if equal, size decision
			if (newSize>bestSize)
			{
				continue;
			} else if (newSize<bestSize)
			{
				newCurr = new CommunityWithStrength(toCommunity, newStrength);
				bestSize = newSize;
				continue;
			}
			
			// if equal, ID decision
			if (toCommunity>newCurr.community)
				continue;
			// new because of same strength but smaller ID 
			newCurr = new CommunityWithStrength(toCommunity, newStrength);
		}
		
		return newCurr;
	}	

	/**
	 * glorified argmax 
	 * 
	 * @param toMap
	 * @param toContainer
	 * @param overrideMap override values in toMap
	 * @param overrideSize override values in toContainer.getSize()
	 * @return
	 */
	CommunityWithStrength getBest(Map<Integer, Double> toMap, 	
								  EdgeContainer toContainer,
							      Map<Integer, Double> overrideMap,
							      Map<Integer, Double> overrideSize)
	{
		
		if (overrideMap!=null || overrideSize!=null)
			throw (new RuntimeException("aua!"));
		
		
		int bestCommunity = -1;
		double bestStrength = -1;
		double bestSize = -1;
		
		if (toMap.size()==0)
			throw (new RuntimeException("getBest called with empty map!"));
		
		Set<Integer> allCommunities = null;
		allCommunities = toMap.keySet();
		
		for (int community: allCommunities)
		{
			double strength = 0;
			strength = toMap.get(community);
			if (strength<bestStrength)
				continue;
			double newSize = toContainer.getSize(community);
			if (strength>bestStrength)
			{
				bestStrength = strength;
				bestCommunity = community;
				bestSize = newSize;
				continue;
			}

			// decide for smaller community
			if (newSize>bestSize)
				continue;
			if (newSize<bestSize)
			{
				bestStrength = strength;
				bestCommunity = community;
				bestSize = newSize;			
			}
			
			// if we are here, current and best community are also of equal strength
			// choose smaller index just to make things deterministic as opposed to dependent
			// on order in Map traversal
			if (bestCommunity>community)
			{
				bestStrength = strength;
				bestCommunity = community;
				bestSize = newSize;			
			}
			
		}
		
		return new CommunityWithStrength(bestCommunity, bestStrength);
	}	
	
	
		
}
