package de.tuberlin.cs.ni.neubauer.mpcd.io;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

public class DataContainer
{
	List<int[]> edges;
	int[] numNodes;
	Map<Integer, Map<Integer, Integer>> newToOld; 
	int numDimensions;
	public Boolean weighted;
	public Boolean renumbered;
	public List<Float> weights;

	public DataContainer(List<int[]> edges)
	{
		this(edges, false, null, false);		
	}
	
	
	public DataContainer(List<int[]> edges, Boolean weighted, List<Float> weights)
	{
		this(edges, weighted, weights, false);		
	}
	
	public DataContainer(List<int[]> edges, Boolean weighted, List<Float> weights, Boolean overrideRenumbering)
	{
		if (edges.size()==0)
			throw (new RuntimeException("empty list of edges passed on to DataContainer"));
		this.weighted = weighted;
		// if edge-tuples contain weight, actual dimension is tuple length - 1
		this.numDimensions = edges.get(0).length;
		this.edges = renumber(edges, overrideRenumbering);
		this.weights = weights;
		
	}
	
	
	/**
	 * renumbers node IDs to range from 0..count(nodes)-1
	 * as a side effect, initializes numNodes, renumbered and newToOld
	 * 
	 * 
	 * @param edges
	 * @param overrideRenumbering override renumbering; only numNodes is initialized
	 * @return
	 */
	public List<int[]> renumber(List<int[]> edges, Boolean overrideRenumbering)
	{
		return renumber(edges, overrideRenumbering, true);
	}
	
	public List<int[]> renumber(List<int[]> edges, Boolean overrideRenumbering, Boolean alwaysRenumber)
	{
		this.numNodes = new int[this.numDimensions];

		Map<Integer, Map<Integer, Integer>> oldToNew = new HashMap<Integer, Map<Integer, Integer>>();
		this.newToOld = new HashMap<Integer, Map<Integer, Integer>>();
		for (int i=0; i<this.numDimensions; i++)
		{
			oldToNew.put(i, new HashMap<Integer, Integer>());
			this.newToOld.put(i, new HashMap<Integer, Integer>());
		}
		
		for (int[] edge: edges)
		{
			for (int i=0; i<this.numDimensions; i++)
			{
				int oldNodeId = edge[i];
				if (oldNodeId==-1)
					continue;
				Map<Integer, Integer> oldToNewI = oldToNew.get(i);
				if (oldToNewI.containsKey(oldNodeId))
					continue;
				oldToNewI.put(oldNodeId, this.numNodes[i]);
				this.newToOld.get(i).put(this.numNodes[i], oldNodeId);
				this.numNodes[i]++;
			}
		}
		this.renumbered = alwaysRenumber;
		if (!overrideRenumbering && !alwaysRenumber)
		{
			for (int i=0; i<this.numDimensions; i++)
				if (this.numNodes[i]-1<Collections.max(oldToNew.get(i).keySet()))
				{
					this.renumbered = true;
					break;
				}
		}
		
		if (!this.renumbered)
		{
			this.newToOld = null;
			return edges;
		}
		List<int[]> newEdges = new ArrayList<int[]>();
		for (int[] edge: edges)
		{
			int[] newEdge = new int[this.numDimensions];
			for (int i=0; i<this.numDimensions; i++)
			{
				if (edge[i]!=-1)
					newEdge[i] = oldToNew.get(i).get(edge[i]);
			}
			newEdges.add(newEdge);
		}
		return newEdges;
	}
	
	public int getNewToOld(int dim, int id)
	{
		if (id==-1)
			return -1;
		if (!this.newToOld.containsKey(dim))
			throw (new RuntimeException(String.format("dim %d not found!?", dim)));
		
		if (!this.newToOld.get(dim).containsKey(id))
			throw (new RuntimeException(String.format("%d.%d not found!", dim, id)));
		
		return this.newToOld.get(dim).get(id);
	}
	

	/**
	 * original data is no longer needed, but other information from this class is
	 * so this method allows the release of the original data for memory savings
	 */
	public void releaseEdges()
	{
		this.edges = null;
	}
	
	
	public int numDimensions()
	{
		return this.numDimensions;
	}

	
	public List<int[]> getEdges()
	{
		return this.edges;
	}
	
	public List<Float> getWeights()
	{
		return this.weights;
	}
	
	
	/**
	 * used to replace new->community mappings to old->community mappings 
	 * @param original
	 * @return
	 */
	public Map<Integer, Integer> translateMap(Map<Integer, Integer> newToCommunity, int dimension)
	{
		if (!this.renumbered)
			return newToCommunity;
		Map<Integer, Integer> newMap = new HashMap<Integer, Integer>();
		Map<Integer, Integer> translation = this.newToOld.get(dimension);
		for (Map.Entry<Integer, Integer> e: newToCommunity.entrySet())
			newMap.put(translation.get(e.getKey()), e.getValue());
		return newMap;
	}
	
	public void dumpNewToOldMap(String filename) throws IOException
	{
		Writer output;
		String newline = System.getProperty("line.separator");
		
		for (int i=0; i<3; i++)
		{
			File file = new File(filename+"."+i);			
			output = new BufferedWriter(new FileWriter(file));
			
			for (Entry<Integer, Integer> e: this.newToOld.get(i).entrySet())
			{
				int newId = e.getKey();
				int oldId = e.getValue();
				output.write(String.format("%d\t%d%s", newId, oldId, newline));
			}
			output.close();
		}
	}

}
