package de.tuberlin.cs.ni.neubauer.mpcd.listeners;

import java.util.Map;

import de.tuberlin.cs.ni.neubauer.mpcd.Optimizer;
import de.tuberlin.cs.ni.neubauer.mpcd.edges.EdgeContainer;
import de.tuberlin.cs.ni.neubauer.mpcd.modularity.DomainState;
import de.tuberlin.cs.ni.neubauer.mpcd.modularity.MurataDomainState;
import de.tuberlin.cs.ni.neubauer.mpcd.modularity.MurataModularityState;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.CommunityWithStrength;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.JoinPair;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.Pair;

public class ModularityCheckListener extends NoShutDownListener
{

	/*
	 * this found three errors (8.1.)
	 * - in ComputeGain (outgoing modularities), switching corr comm bigger -> smaller ended up in the wrong
	 *   branch (the "completely new corr comm" branch), leading to wrong weight/strength values
	 * - ListEdgeContainer.getBestJointCommunityOneEmpty: in the crucial part inside getBest, the original
	 *   and not the added size of the new community was used for computation, so the "joint community gets too big
	 *   and is overtaken by another, similarly well-connected but smaller community as corr. comm" case never
	 *   actually happened
	 * - ListEdgeContainer.getBestOtherCommunity: faulty creation of overrideMap, losing edges of joint community
	 * @see de.tuberlin.cs.ni.neubauer.mpcd.listeners.OptimizationStepListener#onStep(de.tuberlin.cs.ni.neubauer.mpcd.Optimizer, de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.JoinPair)
	 */
	
	@Override
	public Boolean onStep(Optimizer optimizer, JoinPair joinPair) throws ListenerException
	{
		MurataModularityState modularityState = (MurataModularityState)optimizer.modularity;

		double currentModularity = 0;
		final double epsilon = 0.000001;
		
		for (Map.Entry<Integer, DomainState> entry: modularityState.domainStates.entrySet())
		{
			int fromDomain = entry.getKey();
			MurataDomainState domainState = (MurataDomainState)entry.getValue();
			for (Map.Entry<Integer, EdgeContainer> innerEntry: domainState.edges.entrySet())
			{
				int toDomain = innerEntry.getKey();
				EdgeContainer edges = innerEntry.getValue();
				EdgeContainer otherEdges = modularityState.domainStates.get(toDomain).edges.get(fromDomain);
				Map<Integer, Double> modularity = domainState.modularity.get(toDomain);				
				Map<Integer, CommunityWithStrength> correspondingCommunity = domainState.correspondingCommunities.get(toDomain);
				
				for (Map.Entry<Integer, CommunityWithStrength> bestPair: edges.getBestCommunities(otherEdges).entrySet())
				{
					int communityFrom = bestPair.getKey();
					int communityTo = bestPair.getValue().community;
					double strength = bestPair.getValue().strength;
					double a_my = edges.getSize(communityFrom);
					double a_other = otherEdges.getSize(communityTo);
					double norm = edges.norm();
					double m = norm * (strength - norm*a_my*a_other);
					
					double predictedM = modularity.get(communityFrom);

					int predictedCommunityTo = correspondingCommunity.get(communityFrom).community;
					if (communityTo!=predictedCommunityTo)
					{
						String message = String.format("wrong predicted corresponding community in %d for %d.%d! \nright: %d (str=%1.2f, weight=%1.2f)\nwrong predicted: %d (str=%1.2f, realstr=%1.2f, weight=%1.2f)",
								toDomain,
								fromDomain,
								communityFrom,
								communityTo, 
								strength,
								otherEdges.getSize(communityTo),
								predictedCommunityTo,
								correspondingCommunity.get(communityFrom).strength,
								otherEdges.connectionsTo(predictedCommunityTo, new Pair(communityFrom, -1)).smaller_element,
								otherEdges.getSize(predictedCommunityTo)); 
						throw (new RuntimeException(message));
					}
					
					if (Math.abs(m-predictedM)>epsilon)
						throw (new ListenerException(String.format("wrong predicted modularity! %1.9f != %1.9f\ndim %d->%d\nnode %d->%d", m, predictedM, fromDomain, toDomain, communityFrom, communityTo)));
					currentModularity += m;
				}
			}
		}		
		
		return false;
	}

 
}
