package de.tuberlin.cs.ni.neubauer.mpcd.modularity;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import de.tuberlin.cs.ni.neubauer.mpcd.assignments.CommunityAssignment;
import de.tuberlin.cs.ni.neubauer.mpcd.assignments.CommunityAssignments;
import de.tuberlin.cs.ni.neubauer.mpcd.edges.EdgeContainer;
import de.tuberlin.cs.ni.neubauer.mpcd.edges.EdgeContainerFactory;
import de.tuberlin.cs.ni.neubauer.mpcd.gain.GainTrackManager;
import de.tuberlin.cs.ni.neubauer.mpcd.gain.GainTrackerFactory;
import de.tuberlin.cs.ni.neubauer.mpcd.io.DataContainer;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.JoinPair;

public abstract class AbstractModularityState implements ModularityState
{
	
	public Map<Integer, DomainState> domainStates; // TODO shouldn't be public
	int numDimensions;
	CommunityAssignments assignments;
	double currentModularity;
	GainTrackerFactory gainTrackerFactory;
	DataContainer data;
	
	public AbstractModularityState(DataContainer data, EdgeContainerFactory edgeFactory, GainTrackerFactory gainTrackerFactory, Boolean unweightedDU)
	{
		this.gainTrackerFactory = gainTrackerFactory;
		this.numDimensions = data.numDimensions();
		this.data = data;
		/*
		if (this.numDimensions!=3)
			throw (new RuntimeException(String.format("undefined for #partitions %d!=3", this.numDimensions)));
			*/
		this.domainStates = new HashMap<Integer, DomainState>();
		for (int i=0; i<this.numDimensions; i++)
			this.domainStates.put(i, getDomainState(i, this.numDimensions, unweightedDU, edgeFactory));
		initFromEdges(data);
	}	
	
	public int getNumDimensions()
	{
		return this.numDimensions;
	}
	
	public DataContainer getData()
	{
		return this.data;
	}
	
	
	// although some interface methods are implemented, new but non-public
	// abstract methods are defined as well

	/**
	 * 	creates new DomainStates - implementing subclasses of AbstractModularityState
	 *  may wish to subclass DomainState as well 
	 */
	abstract DomainState getDomainState(int i, int partitions, Boolean unweightedDU, EdgeContainerFactory edgeFactory);

	/** init community-wise modularity values in DomainStates */
	abstract void initModularity();
	
	/** update GainTrackers and merge edges (has to be done in-between, regrettably) */
	// TODO check if edge merging can painlessly be done BEFORE updateGain
	abstract void updateGain(JoinPair pair, GainTrackManager manager);
	

	
	@Override
	public CommunityAssignments getCommunityAssignments()
	{
		return this.assignments;
	}


	@Override
	public void join(JoinPair pair, GainTrackManager manager)
	{
		// we're not interested in the actual gain computation which should already be in pair
		// instead we're interested in any metadata generated here
 		computeGain(pair, true);
 		
		// join community assignments
		this.assignments.get(pair.dimension).join(pair.pair);
		
		// so the modularity-specific updateGain can benefit from this information
		updateGain(pair, manager);
		
		// update internal gain counter
		this.currentModularity += pair.gain;
		
		
		// merge edges
	}
	
	@Override
	public double getCurrentModularity()
	{
		return this.currentModularity;
	}
	

	
	/**
	 * initializes edges and communities based on a list of edges
	 * @param edges
	 */
	public void initFromEdges(DataContainer data)
	{
		List<int[]> edges = data.getEdges();
		int l = edges.get(0).length;
		if (l!=this.numDimensions)
		{
			throw (new RuntimeException("undefined for #values!=partitions"));
		}		
		
		// setup finding out the number of elements
		int[] max = new int[this.numDimensions];
		for (int i=0; i<this.numDimensions; i++)
			max[i] = 0;
		
		// store edges between communities (now still nodes)
		int index=0;
		for (int[] values: edges)
		{
			float s = 1;
			if (data.weighted)
				s = data.weights.get(index);
			index++;
			for (int i=0; i<this.numDimensions; i++)
				for (int j=i+1; j<this.numDimensions; j++)
				{
					int vali = values[i];
					int valj = values[j];
					if (vali==-1 || valj==-1)
						continue;
					if (vali>max[i])
						max[i] = vali;
					if (valj>max[j])
						max[j] = valj;
					
					getEdges(i,j).add(vali, valj, s);
					getEdges(j,i).add(valj, vali, s);
				}
		}		
		
		// store number of elements
		for (int i=0; i<this.numDimensions; i++)
			this.domainStates.get(i).setNumNodes(max[i]+1);
		
		// now modularity can be computed
		initModularity();

		// create a community for each node		
		this.assignments = new CommunityAssignments(this.numDimensions, 0, this.currentModularity);
		for (int i=0; i<this.numDimensions; i++)
		{
			DomainState ds = this.domainStates.get(i);
			Collection<Integer> nodes = new HashSet<Integer>();
			for (EdgeContainer e: ds.edges.values())
			{
				nodes.addAll(e.allFromCommunities());
			}
			this.assignments.put(i, new CommunityAssignment(nodes));	
		}
	}	
	
	
	/**
	 * a shortcut function to get to the right edge container quickly
	 * @param from the partition edges are coming from
	 * @param to the partition edges are going to
	 * @return an EdgeContainer
	 */
	public final EdgeContainer getEdges(int from, int to)
	{
		return this.domainStates.get(from).edges.get(to);
	}	

}
