package de.tuberlin.cs.ni.neubauer.mpcd.modularity;

import java.util.ArrayList;
import java.util.List;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import de.tuberlin.cs.ni.neubauer.mpcd.edges.EdgeContainer;
import de.tuberlin.cs.ni.neubauer.mpcd.edges.EdgeContainerFactory;
import de.tuberlin.cs.ni.neubauer.mpcd.gain.DefaultGainTracker;
import de.tuberlin.cs.ni.neubauer.mpcd.gain.GainTrackManager;
import de.tuberlin.cs.ni.neubauer.mpcd.gain.GainTracker;
import de.tuberlin.cs.ni.neubauer.mpcd.gain.GainTrackerFactory;
import de.tuberlin.cs.ni.neubauer.mpcd.io.DataContainer;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.CommunityWithStrength;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.DoublePair;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.JoinPair;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.Pair;



public class MurataModularityState extends AbstractModularityState
{
	
	private GainMetadata lastJoinMetadata;
	public int[] pairsChanged;
	public int[] pairsComputed;
	public int[] pairsMax;
	public Map<Integer, List<Pair>> pairs;
	
	public MurataModularityState(DataContainer data, 
			EdgeContainerFactory edgeFactory, 
			GainTrackerFactory gainTrackerFactory, 
			Boolean unweightedDU)
	{
		super(data, edgeFactory, gainTrackerFactory, unweightedDU);
	}	


	@Override
	public GainTrackManager getGainTrackManager()
	{
		GainTrackManager gtm = this.gainTrackerFactory.getGainTrackManager(this.numDimensions);
		for (int dimension=0; dimension<this.numDimensions; dimension++)
		{
			Map<Integer, Map<Integer, Double>> map = new HashMap<Integer, Map<Integer, Double>>();
			
			int numNodes = this.domainStates.get(dimension).numNodes;
			for (int smaller = 0; smaller < numNodes-1; smaller++)
			{
				if (gtm.verboseOutput())
					System.out.println(String.format("%d: %d/%d", dimension, smaller+1, numNodes));
				Map<Integer, Double> innerMap = new HashMap<Integer, Double>();
				map.put(smaller, innerMap);
				for (int bigger = smaller+1; bigger < numNodes; bigger++)
				{
					innerMap.put(bigger, computeGain(new JoinPair(new Pair(smaller, bigger), dimension), false));
				}
			}
			GainTracker gt = new DefaultGainTracker(map);
			gtm.put(dimension, gt);
		}
		return gtm;
	}

	@Override
	void initModularity()
	{
		this.currentModularity = 0;
		
		for (Map.Entry<Integer, DomainState> entry: this.domainStates.entrySet())
		{
			int fromDomain = entry.getKey();
			MurataDomainState domainState = (MurataDomainState)entry.getValue();
			for (Map.Entry<Integer, EdgeContainer> innerEntry: domainState.edges.entrySet())
			{
				int toDomain = innerEntry.getKey();
				EdgeContainer edges = innerEntry.getValue();
				Map<Integer, Double> modularity = domainState.modularity.get(toDomain);
				Map<Integer, CommunityWithStrength> correspondingCommunity = domainState.correspondingCommunities.get(toDomain);
				EdgeContainer otherEdges = this.domainStates.get(toDomain).edges.get(fromDomain);
				for (Map.Entry<Integer, CommunityWithStrength> bestPair: edges.getBestCommunities(otherEdges).entrySet())
				{
					int communityFrom = bestPair.getKey();
					int communityTo = bestPair.getValue().community;
					double strength = bestPair.getValue().strength;
					double a_my = edges.getSize(communityFrom);
					double a_other = otherEdges.getSize(communityTo);
					double norm = edges.norm();
					double m = norm * (strength - norm*a_my*a_other);
					modularity.put(communityFrom, m);
					correspondingCommunity.put(communityFrom, bestPair.getValue());
					this.currentModularity += m;
				}
			}
		}		
	}

	@Override
	DomainState getDomainState(int i, int partitions, Boolean unweightedDU, EdgeContainerFactory edgeFactory)
	{
		return new MurataDomainState(i, partitions, unweightedDU, edgeFactory);
	}
	
	
	class GainMetadata
	{
		// dim from -> dim to -> community -> modularity
		public Map<Integer, Map<Integer, Map<Integer, Double>>> newModularity;
		
		// dim from -> dim to -> community -> corresponding
		public Map<Integer, Map<Integer, Map<Integer, CommunityWithStrength>>> newCorrespondingCommunities;
		
		public Set<Integer> affectedCommunities;
		
		public GainMetadata(JoinPair pair)
		{
			this.newCorrespondingCommunities = new HashMap<Integer, Map<Integer, Map<Integer, CommunityWithStrength>>>();
			this.newModularity = new HashMap<Integer, Map<Integer, Map<Integer, Double>>>();
			this.affectedCommunities = new HashSet<Integer>();
			this.affectedCommunities.add(pair.pair.smaller_element);
		}
		
		
		public void storeCorrespondingCommunities (int from, int to, int community, CommunityWithStrength correspondingCommunity)
		{
			Map<Integer, Map<Integer, CommunityWithStrength>> dimensionMap = null;
			if (!this.newCorrespondingCommunities.containsKey(from))
			{
				dimensionMap = new HashMap<Integer, Map<Integer, CommunityWithStrength>>();
				this.newCorrespondingCommunities.put(from, dimensionMap);
			}
			else
			{
				dimensionMap = this.newCorrespondingCommunities.get(from);
			}
			
			Map<Integer, CommunityWithStrength> innerMap = null;
			if (!dimensionMap.containsKey(to))
			{
				innerMap = new HashMap<Integer, CommunityWithStrength>();
				dimensionMap.put(to, innerMap);
			}
			else
			{
				innerMap = dimensionMap.get(to);
			}
			innerMap.put(community, correspondingCommunity);
		}
		
		public Map<Integer, CommunityWithStrength> getCorrespondingCommunities(int from, int to)
		{
			if (!this.newCorrespondingCommunities.containsKey(from))
				return new HashMap<Integer, CommunityWithStrength>();
			Map<Integer, Map<Integer, CommunityWithStrength>> map = this.newCorrespondingCommunities.get(from);
			if (!map.containsKey(to))
				return new HashMap<Integer, CommunityWithStrength>();
			return map.get(to);
		}
		
		public void storeNewModularity(int from, int to, int community, double modularity)
		{
			Map<Integer, Map<Integer, Double>> dimensionMap = null;
			if (!this.newModularity.containsKey(from))
			{
				dimensionMap = new HashMap<Integer, Map<Integer, Double>>();
				this.newModularity.put(from, dimensionMap);
			}
			else
			{
				dimensionMap = this.newModularity.get(from);
			}
			
			Map<Integer, Double> innerMap = null;
			if (!dimensionMap.containsKey(to))
			{
				innerMap = new HashMap<Integer, Double>();
				dimensionMap.put(to, innerMap);
			}
			else
			{
				innerMap = dimensionMap.get(to);
			}
			innerMap.put(community, modularity);			
		}
		
		public Map<Integer, Double> getModularity(int from, int to)
		{
			if (!this.newModularity.containsKey(from))
				return new HashMap<Integer, Double>();
			Map<Integer, Map<Integer, Double>> map = this.newModularity.get(from);
			if (!map.containsKey(to))
				return new HashMap<Integer, Double>();
			return map.get(to);
			
		}
		
		public void addAffectedCommunity(int community)
		{
			this.affectedCommunities.add(community);
		}
	}

	
	double gainChange(int otherCommunity, 
			EdgeContainer otherEdgesToMe, 
			double newShared,
			double norm, // equal in both directions
			double myWeight)
			
	{
		double otherWeight = otherEdgesToMe.getSize(otherCommunity);
		double newModularity = norm * (newShared - norm*myWeight*otherWeight);
		return newModularity;
	}
	

	
	@Override
	public double computeGain(JoinPair joinPair, Boolean computeMetadata)
	{
		
		int dimension = joinPair.dimension;
		int smaller_element = joinPair.pair.smaller_element;
		int bigger_element = joinPair.pair.bigger_element;
		MurataDomainState myDomainState = (MurataDomainState)this.domainStates.get(dimension);
		
		
		if (computeMetadata)
		{
			this.lastJoinMetadata = new GainMetadata(joinPair);
		}
		double gain = 0.0;

		// gain change concerning outgoing modularities
		for (Map.Entry<Integer, Map<Integer, Double>> entry: myDomainState.modularity.entrySet())
		{
			int otherDimension = entry.getKey();
			Map<Integer, Double> modularity = entry.getValue();
			EdgeContainer edgesToOther = myDomainState.edges.get(otherDimension);
			
			// start by subtracting the current modularities going out from both communities
			gain = gain - modularity.get(smaller_element)
				        - modularity.get(bigger_element);
			
			// compute modularity of joint community, depending on new corresponding community
			EdgeContainer otherEdges = this.domainStates.get(otherDimension).edges.get(dimension);
			CommunityWithStrength newCorresponding =  
				myDomainState.edges.get(otherDimension).getBestOtherCommunity(joinPair.pair, otherEdges,
						myDomainState.correspondingCommunities.get(otherDimension));
			double sharedEdges = newCorresponding.strength;
			int other_community = newCorresponding.community;
			
			double myWeight = edgesToOther.getSize(smaller_element) + edgesToOther.getSize(bigger_element);
			double otherWeight = otherEdges.getSize(other_community);
			double norm = edgesToOther.norm();
			double newModularity = norm * (sharedEdges - norm * myWeight * otherWeight);
			gain += newModularity;
			
			// bookkeeping if we're interested
			if (computeMetadata)
			{
				this.lastJoinMetadata.storeCorrespondingCommunities(dimension, otherDimension, smaller_element, newCorresponding);
				this.lastJoinMetadata.storeNewModularity(dimension, otherDimension, smaller_element, newModularity);
			}
		}
		
		// gain change in *incoming* modularities
		for (Map.Entry<Integer, DomainState> entry: this.domainStates.entrySet())
		{
			int otherDimension = entry.getKey();
			if (otherDimension==dimension)
				continue;

			MurataDomainState otherDomainState = (MurataDomainState)entry.getValue();
			Map<Integer, CommunityWithStrength> otherCorresponding = otherDomainState.correspondingCommunities.get(dimension);
			EdgeContainer otherEdgesToMe = otherDomainState.edges.get(dimension);
			Map<Integer, Double> otherModularity = otherDomainState.modularity.get(dimension);
			
			EdgeContainer edgesToOther = myDomainState.edges.get(otherDimension);
			double norm = edgesToOther.norm();
			double myWeight = edgesToOther.getSize(smaller_element) + edgesToOther.getSize(bigger_element);
			

			for (int otherCommunity: otherEdgesToMe.allFromCommunities())
			{
				CommunityWithStrength currentCorrCommWS = otherCorresponding.get(otherCommunity);
				int currentCorrComm = currentCorrCommWS.community;
				double currentSharedEdges = currentCorrCommWS.strength;
				DoublePair shared = otherEdgesToMe.connectionsTo(otherCommunity, joinPair.pair);
				double newSharedEdges = shared.smaller_element + shared.bigger_element;
				double oldModularity = otherModularity.get(otherCommunity);
				
				// first case: one of the to-be-joint nodes already correspond
				// then joint community will continue to correspond, although modularity value will change
				// (unless it doesn't get better but bigger than competitor)
				if (joinPair.pair.contains(currentCorrComm)) 
				{
                    // if we don't gain (connections is 0 for node_i or node_j)
                    // although the other one is currently the corresponding community
					if (shared.bigger_element==0 || shared.smaller_element==0) 
					{
						CommunityWithStrength newCorrCommWS = otherEdgesToMe.getBestJointCommunityOneEmpty(otherCommunity, joinPair.pair, edgesToOther, newSharedEdges);
						int newCorrComm = newCorrCommWS.community;						
						if (!joinPair.pair.contains(newCorrComm))
						{
							// here, we handle the case of a changed corresponding community
							// because the present one (one of the joints) has become too large without improvement							
							double newModularity = gainChange(otherCommunity,
									otherEdgesToMe,
									newCorrCommWS.strength,
									norm,
									edgesToOther.getSize(newCorrComm));
							gain += newModularity - oldModularity;
							if (computeMetadata)
							{
								this.lastJoinMetadata.storeCorrespondingCommunities(otherDimension, dimension, otherCommunity, newCorrCommWS);
								this.lastJoinMetadata.storeNewModularity(otherDimension, dimension, otherCommunity, newModularity);
								this.lastJoinMetadata.addAffectedCommunity(newCorrComm);
							}
							continue;
						}
					}
					
					// if we are here, one of two's still corresponding, but modularity may have changed
					double newModularity = gainChange(otherCommunity,
							otherEdgesToMe,
							newSharedEdges,
							norm,
							myWeight);			
					gain += newModularity - oldModularity;
					if (computeMetadata)
					{
						this.lastJoinMetadata.storeCorrespondingCommunities(otherDimension, dimension, otherCommunity, new CommunityWithStrength(smaller_element, newSharedEdges));
						this.lastJoinMetadata.storeNewModularity(otherDimension, dimension, otherCommunity, newModularity);
						this.lastJoinMetadata.addAffectedCommunity(smaller_element);
						this.lastJoinMetadata.addAffectedCommunity(bigger_element);
					}
					continue;
				}
				
				// second case: communities may be connected or not, but are not corresponding
				// let's see if the joint community becomes corresponding
				
				// we need to share more edges than current to take over correspondence
				if (newSharedEdges < currentSharedEdges)
					continue;
				// or if equal fulfill other criteria
				if (newSharedEdges == currentSharedEdges)
				{
					double currCommWeight = edgesToOther.getSize(currentCorrComm);
					// not be heavier
					if (myWeight>currCommWeight)
						continue;
					else if (myWeight==currCommWeight)
					{
						// or if equally heavy, have a lower ordinal number 
						if (smaller_element > currentCorrComm)
							continue; // tie-breaking strategy from getBestCommunity nachgebaut
					}
				}
				// if we arrive here, we take over correspondence!
				double newModularity = gainChange(otherCommunity,
						otherEdgesToMe,
						newSharedEdges,
						norm,
						myWeight);			
				gain += newModularity - oldModularity;
				if (computeMetadata)
				{
					this.lastJoinMetadata.storeCorrespondingCommunities(otherDimension, dimension, otherCommunity, new CommunityWithStrength(smaller_element, newSharedEdges));
					this.lastJoinMetadata.storeNewModularity(otherDimension, dimension, otherCommunity, newModularity);
					this.lastJoinMetadata.addAffectedCommunity(currentCorrComm);
				}				
			} // otherCommunity
		} // otherDimension
				
		return gain;
	}
	
	void updateModularitiesAndCommunities(JoinPair joinPair, GainTrackManager gtm)
	{
		
	}

	@Override
	void updateGain(JoinPair joinPair, GainTrackManager gtm)
	{
		int dimension = joinPair.dimension;
		int bigger_element = joinPair.pair.bigger_element;
		
		
		Map<Integer, Set<Integer>> affectedCommunities = new HashMap<Integer, Set<Integer>>();
		Set<Integer> myAffected = this.lastJoinMetadata.affectedCommunities;
		affectedCommunities.put(dimension, myAffected);
		MurataDomainState myDomainState = (MurataDomainState)this.domainStates.get(dimension);

		
		/** first step: find affected communities */ 
		// first, all other communities connected to affected ones from changed domain are affected
		for (int otherDimension: this.domainStates.keySet())
		{
			if (otherDimension==dimension)
				continue;
			EdgeContainer myEdgesToOther = myDomainState.edges.get(otherDimension); 
			Set<Integer> affected = new HashSet<Integer>();
			// what's returned is an immutable keyset
			affected.addAll(myEdgesToOther.allToCommunities(joinPair.pair.smaller_element));
			Set<Integer> alsoAffected = myEdgesToOther.allToCommunities(joinPair.pair.bigger_element); 
			affected.addAll(alsoAffected);
			affectedCommunities.put(otherDimension, affected);
		}
			
		// second, all communities in changed domain connected to those affected from the previous step
		// are also affected
		for (int otherDimension: this.domainStates.keySet())
		{
			if (otherDimension==dimension)
				continue;					
			EdgeContainer otherEdgesToMe = this.domainStates.get(otherDimension).edges.get(dimension);
			for (int otherCommunity: affectedCommunities.get(otherDimension))
			{
				for (int myDomainCommunity: otherEdgesToMe.allToCommunities(otherCommunity))
					myAffected.add(myDomainCommunity);
			}
		}
		
		/** second step: update modularities and corresponding communities */ 
		for (int otherDimension: this.domainStates.keySet())
		{		
			if (otherDimension==dimension)
				continue;	

			// in joint dimension
			Map<Integer, Double> myModularity = myDomainState.modularity.get(otherDimension);
			Map<Integer, CommunityWithStrength> myCorrComms = myDomainState.correspondingCommunities.get(otherDimension);
			myModularity.remove(bigger_element);
			myCorrComms.remove(bigger_element);
			for (Map.Entry<Integer, Double> e: 
				this.lastJoinMetadata.newModularity.get(dimension).get(otherDimension).entrySet())
			{
				int community = e.getKey();
				double newModularity = e.getValue();
				myModularity.put(community, newModularity);
				myAffected.add(community);
			}
			
			for (Map.Entry<Integer, CommunityWithStrength> e: 
				this.lastJoinMetadata.newCorrespondingCommunities.get(dimension).get(otherDimension).entrySet())
			{
				int communityFrom = e.getKey();
				myCorrComms.put(communityFrom, e.getValue());
				myAffected.add(communityFrom);
			}
			
			// in other dimension
			MurataDomainState otherDomainState = (MurataDomainState)this.domainStates.get(otherDimension);
			Map<Integer, Double> otherModularity = otherDomainState.modularity.get(dimension);
			Map<Integer, CommunityWithStrength> otherCorrComms = otherDomainState.correspondingCommunities.get(dimension);
			Set<Integer> otherAffected = affectedCommunities.get(otherDimension); 
			Set<Integer> check = new HashSet<Integer>();
			
			for (Map.Entry<Integer, Double> e: this.lastJoinMetadata.getModularity(otherDimension, dimension).entrySet())
			{
				int community = e.getKey();
				double newModularity = e.getValue();
				otherModularity.put(community, newModularity);
				if (!otherAffected.contains(community))
				{
					otherAffected.add(community);
					check.add(community);
				}				
			}
			
			for (Map.Entry<Integer, CommunityWithStrength> e: this.lastJoinMetadata.getCorrespondingCommunities(otherDimension, dimension).entrySet())
			{
				int communityFrom = e.getKey();
				otherCorrComms.put(communityFrom, e.getValue());
				if (!otherAffected.contains(communityFrom))
				{
					otherAffected.add(communityFrom);
					check.add(communityFrom);
				}
			}
			
			// now we have potentially additional affected communities in other regions, 
			// transmit changes back to joint dimension
			EdgeContainer otherEdgesToMe = otherDomainState.edges.get(dimension);
			for (int otherCommunity: check)
			{
				for (int myDomainCommunity: otherEdgesToMe.allToCommunities(otherCommunity))
					myAffected.add(myDomainCommunity);
			}
		}
		
		/** third step: join edges */
		
		for (DomainState domainstate: this.domainStates.values())
			domainstate.join(joinPair.dimension, joinPair.pair);
		
		/** fourth step: finally update gain */
		gtm.get(dimension).removeElement(bigger_element);
		myAffected.remove(bigger_element);
		this.pairsComputed = new int[this.numDimensions];
		this.pairsChanged = new int[this.numDimensions];
		this.pairsMax = new int[this.numDimensions];
		this.pairs = new HashMap<Integer, List<Pair>>();
		for (int updateDimension: this.domainStates.keySet())
		{
			ArrayList<Pair> p = new ArrayList<Pair>();
			this.pairs.put(updateDimension, p);
			GainTracker gt = gtm.get(updateDimension);
			Set<Integer> affected = affectedCommunities.get(updateDimension);
			Set<Integer> all = this.assignments.get(updateDimension).byCommunity().keySet();
			int m = all.size();
			m = (m*(m-1))/2;
			this.pairsMax[updateDimension] = m;
			for (int otherCommunity: all)
			{
				Boolean otherIsAffected = affected.contains(otherCommunity);
				for (int affectedCommunity: affected)
				{
					if (otherCommunity == affectedCommunity)
						continue;
					if (otherIsAffected && affectedCommunity<otherCommunity)
						continue;
					Pair pair = new Pair(Math.min(otherCommunity, affectedCommunity),
							Math.max(otherCommunity, affectedCommunity));
					JoinPair jp = new JoinPair(pair, updateDimension);
					double newGain = computeGain(jp, false);
					
					
					this.pairsComputed[updateDimension] += 1;
					if (gt.getPairGain(pair)!=newGain)
					{
						this.pairsChanged[updateDimension] += 1;
						gt.setPair(pair, newGain);
						p.add(pair);
					}
				}
			}
		}
		
		if (dimension==0 && joinPair.pair.smaller_element==41 && joinPair.pair.bigger_element==54)
		{
			for (Pair p: this.pairs.get(0))
			{
				System.out.println(p);
			}
			throw (new RuntimeException("x"));
		}
	}
	
	
}
