package de.tuberlin.cs.ni.neubauer.mpcd.modularity;


import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import de.tuberlin.cs.ni.neubauer.mpcd.edges.EdgeContainer;
import de.tuberlin.cs.ni.neubauer.mpcd.edges.EdgeContainerFactory;
import de.tuberlin.cs.ni.neubauer.mpcd.gain.GainTrackManager;
import de.tuberlin.cs.ni.neubauer.mpcd.gain.GainTracker;
import de.tuberlin.cs.ni.neubauer.mpcd.gain.GainTrackerFactory;
import de.tuberlin.cs.ni.neubauer.mpcd.io.DataContainer;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.CommunityWithStrength;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.DoublePair;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.JoinPair;
import de.tuberlin.cs.ni.neubauer.mpcd.smalltypes.Pair;

public class OptimizedMurataModularityState extends MurataModularityState 
{
	
	public int[] connectedPairsComputed;
	public int[] connectedPairsChanged;
	public int[] affectedPairsComputed;
	public int[] affectedPairsChanged;
	
	public OptimizedMurataModularityState(DataContainer data, 
				EdgeContainerFactory edgeFactory, 
				GainTrackerFactory gainTrackerFactory, 
				Boolean unweightedDU)
	{
		super(data, edgeFactory, gainTrackerFactory, unweightedDU);
	}
	
	public OptimizedMurataModularityGainMetaData lastJoinOptMetadata;
	
	
	@Override
	public double computeGain(JoinPair joinPair, Boolean computeMetadata)
	{
		
		int dimension = joinPair.dimension;
		int smaller_element = joinPair.pair.smaller_element;
		int bigger_element = joinPair.pair.bigger_element;
		MurataDomainState myDomainState = (MurataDomainState)this.domainStates.get(dimension);
		
		
		if (computeMetadata)
		{
			this.lastJoinOptMetadata = new OptimizedMurataModularityGainMetaData(joinPair, this.numDimensions);
		}
		double gain = 0.0;

		// gain change concerning outgoing modularities
		for (Map.Entry<Integer, Map<Integer, Double>> entry: myDomainState.modularity.entrySet())
		{
			int otherDimension = entry.getKey();
			Map<Integer, Double> modularity = entry.getValue();
			EdgeContainer edgesToOther = myDomainState.edges.get(otherDimension);
			
			// start by subtracting the current modularities going out from both communities
			gain = gain - modularity.get(smaller_element)
				        - modularity.get(bigger_element);
			
			// compute modularity of joint community, depending on new corresponding community
			EdgeContainer otherEdges = this.domainStates.get(otherDimension).edges.get(dimension);
			CommunityWithStrength newCorresponding =  
				myDomainState.edges.get(otherDimension).getBestOtherCommunity(joinPair.pair, otherEdges,
						myDomainState.correspondingCommunities.get(otherDimension));
			
			double myWeight = edgesToOther.getSize(smaller_element) + edgesToOther.getSize(bigger_element);
			double otherWeight = otherEdges.getSize(newCorresponding.community);
			double norm = edgesToOther.norm();
			double newModularity = norm * (newCorresponding.strength - norm * myWeight * otherWeight);
			gain += newModularity;
			
			// bookkeeping if we're interested
			if (computeMetadata)
			{
				this.lastJoinOptMetadata.storeCorrespondingCommunities(dimension, otherDimension, smaller_element, newCorresponding);
				this.lastJoinOptMetadata.storeNewModularity(dimension, otherDimension, smaller_element, newModularity);
				HashSet<Integer> otherCorresponding = new HashSet<Integer>();
				
				// previous or current corresponding communities in other dimension
				otherCorresponding.add(newCorresponding.community);
				otherCorresponding.add(myDomainState.correspondingCommunities.get(otherDimension).get(smaller_element).community);
				otherCorresponding.add(myDomainState.correspondingCommunities.get(otherDimension).get(bigger_element).community);
				this.lastJoinOptMetadata.addOtherAffectedCommunities(otherCorresponding, otherDimension);
			}
		}
		
		// gain change in *incoming* modularities
		for (Map.Entry<Integer, DomainState> entry: this.domainStates.entrySet())
		{
			int otherDimension = entry.getKey();
			if (otherDimension==dimension)
				continue;

			MurataDomainState otherDomainState = (MurataDomainState)entry.getValue();
			Map<Integer, CommunityWithStrength> otherCorresponding = otherDomainState.correspondingCommunities.get(dimension);
			EdgeContainer otherEdgesToMe = otherDomainState.edges.get(dimension);
			Map<Integer, Double> otherModularity = otherDomainState.modularity.get(dimension);
			
			EdgeContainer edgesToOther = myDomainState.edges.get(otherDimension);
			double norm = edgesToOther.norm();
			double myWeight = edgesToOther.getSize(smaller_element) + edgesToOther.getSize(bigger_element);
			

			for (int otherCommunity: otherEdgesToMe.allFromCommunities())
			{
				CommunityWithStrength currentCorrCommWS = otherCorresponding.get(otherCommunity);
				int currentCorrComm = currentCorrCommWS.community;
				double currentSharedEdges = currentCorrCommWS.strength;
				DoublePair shared = otherEdgesToMe.connectionsTo(otherCommunity, joinPair.pair);
				double newSharedEdges = shared.smaller_element + shared.bigger_element;
				double oldModularity = otherModularity.get(otherCommunity);
				
				// first case: one of the to-be-joint nodes already correspond
				// then joint community will continue to correspond, although modularity value will change
				// (unless it doesn't get better but bigger than competitor)
				if (joinPair.pair.contains(currentCorrComm)) 
				{
                    // if we don't gain (connections is 0 for node_i or node_j)
                    // although the other one is currently the corresponding community
					if (shared.bigger_element==0 || shared.smaller_element==0) 
					{
						CommunityWithStrength newCorrCommWS = otherEdgesToMe.getBestJointCommunityOneEmpty(otherCommunity, joinPair.pair, edgesToOther, newSharedEdges);
						int newCorrComm = newCorrCommWS.community;						
						if (!joinPair.pair.contains(newCorrComm))
						{
							// here, we handle the case of a changed corresponding community
							// because the present one (one of the joints) has become too large without improvement							
							double newModularity = gainChange(otherCommunity,
									otherEdgesToMe,
									newCorrCommWS.strength,
									norm,
									edgesToOther.getSize(newCorrComm));
							gain += newModularity - oldModularity;
							if (computeMetadata)
							{
								this.lastJoinOptMetadata.storeCorrespondingCommunities(otherDimension, dimension, otherCommunity, newCorrCommWS);
								this.lastJoinOptMetadata.storeNewModularity(otherDimension, dimension, otherCommunity, newModularity);
								this.lastJoinOptMetadata.addAffectedCommunity(newCorrComm);
								this.lastJoinOptMetadata.addOtherAffectedCommunity(otherCommunity, otherDimension);
							}
							continue;
						}
					}
					
					// if we are here, one of two's still corresponding, but modularity may have changed
					double newModularity = gainChange(otherCommunity,
							otherEdgesToMe,
							newSharedEdges,
							norm,
							myWeight);			
					gain += newModularity - oldModularity;
					if (computeMetadata)
					{
						this.lastJoinOptMetadata.storeCorrespondingCommunities(otherDimension, dimension, otherCommunity, new CommunityWithStrength(smaller_element, newSharedEdges));
						this.lastJoinOptMetadata.storeNewModularity(otherDimension, dimension, otherCommunity, newModularity);
						this.lastJoinOptMetadata.addAffectedCommunity(smaller_element);
						this.lastJoinOptMetadata.addAffectedCommunity(bigger_element);
						this.lastJoinOptMetadata.addOtherAffectedCommunity(otherCommunity, otherDimension);
					}
					continue;
				}
				
				// second case: communities may be connected or not, but are not corresponding
				// let's see if the joint community becomes corresponding
				
				// we need to share more edges than current to take over correspondence
				if (newSharedEdges < currentSharedEdges)
					continue;
				// or if equal fulfill other criteria
				if (newSharedEdges == currentSharedEdges)
				{
					double currCommWeight = edgesToOther.getSize(currentCorrComm);
					// not be heavier
					if (myWeight>currCommWeight)
						continue;
					else if (myWeight==currCommWeight)
					{
						// or if equally heavy, have a lower ordinal number 
						if (smaller_element > currentCorrComm)
							continue; // tie-breaking strategy from getBestCommunity nachgebaut
					}
				}
				// if we arrive here, we take over correspondence!
				double newModularity = gainChange(otherCommunity,
						otherEdgesToMe,
						newSharedEdges,
						norm,
						myWeight);			
				gain += newModularity - oldModularity;
				if (computeMetadata)
				{
					this.lastJoinOptMetadata.storeCorrespondingCommunities(otherDimension, dimension, otherCommunity, new CommunityWithStrength(smaller_element, newSharedEdges));
					this.lastJoinOptMetadata.storeNewModularity(otherDimension, dimension, otherCommunity, newModularity);
					this.lastJoinOptMetadata.addAffectedCommunity(currentCorrComm);
					this.lastJoinOptMetadata.addOtherAffectedCommunity(otherCommunity, otherDimension);
				}				
			} // otherCommunity
		} // otherDimension
				
		return gain;
	}

	
	void performUpdates(JoinPair joinPair)
	{
		int dimension = joinPair.dimension;
		int bigger_element = joinPair.pair.bigger_element;
		MurataDomainState myDomainState = (MurataDomainState)this.domainStates.get(dimension);		
		
		for (int otherDimension: this.domainStates.keySet())
		{		
			if (otherDimension==dimension)
				continue;	

			// in joint dimension, i.e. outgoing modularity wrt to otherDimension
			Map<Integer, Double> myModularity = myDomainState.modularity.get(otherDimension);
			Map<Integer, CommunityWithStrength> myCorrComms = myDomainState.correspondingCommunities.get(otherDimension);
			myModularity.remove(bigger_element);
			myCorrComms.remove(bigger_element);
			for (Map.Entry<Integer, Double> e: 
				this.lastJoinOptMetadata.newModularity.get(dimension).get(otherDimension).entrySet())
			{
				int community = e.getKey();
				double newModularity = e.getValue();
				myModularity.put(community, newModularity);
			}
			
			for (Map.Entry<Integer, CommunityWithStrength> e: 
				this.lastJoinOptMetadata.newCorrespondingCommunities.get(dimension).get(otherDimension).entrySet())
			{
				int communityFrom = e.getKey();
				myCorrComms.put(communityFrom, e.getValue());
			}
			
			// in other dimension, i.e. otherDimension's outgoing modularity incoming at current dimension
			MurataDomainState otherDomainState = (MurataDomainState)this.domainStates.get(otherDimension);
			Map<Integer, Double> otherModularity = otherDomainState.modularity.get(dimension);
			Map<Integer, CommunityWithStrength> otherCorrComms = otherDomainState.correspondingCommunities.get(dimension);
			
			for (Map.Entry<Integer, Double> e: this.lastJoinOptMetadata.getModularity(otherDimension, dimension).entrySet())
			{
				int community = e.getKey();
				double newModularity = e.getValue();
				otherModularity.put(community, newModularity);
			}
			
			for (Map.Entry<Integer, CommunityWithStrength> e: this.lastJoinOptMetadata.getCorrespondingCommunities(otherDimension, dimension).entrySet())
			{
				int communityFrom = e.getKey();
				otherCorrComms.put(communityFrom, e.getValue());
			}			
		}		
		
	}
	
	
	/**
	 *  in contrast to the previous version, we differentiate between affected and connected.
	 *  
	 *  Affected is any community that 
	 *  - is or was corresponding to one of the joint
	 *  communities (for the other dimension) INCOMING
	 *  
	 *  or
	 *  
	 *  - is corresponding to one of those communities (back in "my" dimension)
	 *  
	 *  Affected communities necessarily (?) change their modularity value and any pair
	 *  containing one of them must be recomputed
	 *  
	 *  Connected is any community that
	 *  - is connected to an affected community 
	 *  in the other dimension (for "my" dimension) INCOMING
	 *  
	 *  or
	 *  
	 *  - is connected to one of the joint communities
	 *  in "my" dimension (for other dimension) OUTGOING
	 *  
	 */	
	void updateGain(JoinPair joinPair, GainTrackManager gtm)
	{
		int dimension = joinPair.dimension;
		int biggerElement = joinPair.pair.bigger_element;
		int smallerElement = joinPair.pair.smaller_element;
		MurataDomainState myDomainState = (MurataDomainState)this.domainStates.get(dimension);		
		
		Set<Integer> myAffected = this.lastJoinOptMetadata.affectedCommunities;
		Map<Integer, Set<Integer>> otherAffected = this.lastJoinOptMetadata.otherAffectedCommunities;
		
		/** first step: compute affected and connected communities */
		// otherDimension -> set of nodes in other dimension connected to my affected communities
		Map<Integer, Set<Integer>> otherConnected = new HashMap<Integer, Set<Integer>>();
		// otherDimension -> set of nodes in my dimension connected to affected communities in other dimension
		Map<Integer, Set<Integer>> myConnected = new HashMap<Integer, Set<Integer>>();
		
		// compute connected communities
		for (int otherDimension: otherAffected.keySet())
		{
			EdgeContainer myEdgesToOther = myDomainState.edges.get(otherDimension); 
			EdgeContainer otherEdgesToMe = this.domainStates.get(otherDimension).edges.get(dimension);			
			
			Map<Integer, CommunityWithStrength> corrCommOther = 
				((MurataDomainState)this.domainStates.get(otherDimension)).correspondingCommunities.get(dimension);	
			
			// test: statt myAffected nur joinPair! (chocho)
			Set<Integer> onlyPair = new HashSet<Integer>(2);
			onlyPair.add(joinPair.pair.smaller_element);
			onlyPair.add(joinPair.pair.bigger_element);
			
			
			Set<Integer> other = myEdgesToOther.getConnectedFromAny(onlyPair); // myAffected
			
			// these guys here are highly annoying.
			// any community in OWN that's connected to any community in OTHER that's connected to joint pair
			// gets AFFECTED status, even though they are ONLY relevant if they stand a chance of taking 
			// correspondence status from or losing it to the new joint community
			// maybe I should monitor those more closely to see if this really pays out
			// have to find the nice little test set which I used to see if the error pops up again if I do
			// (because that's where I first found the need)
			for (int otherCommunity: other)
			{
				// ok it seems this is only required if annoying guy and join-pair have identical number of 
				// nodes. if this is true, we may actually get rid of the final 10% slack
				int currentCorrComm = corrCommOther.get(otherCommunity).community;
				if (currentCorrComm==smallerElement || currentCorrComm==biggerElement)
					continue;
				
				double connSmaller = myEdgesToOther.getStrength(smallerElement, otherCommunity);
				double connBigger = myEdgesToOther.getStrength(biggerElement, otherCommunity);
				double connCorrComm = myEdgesToOther.getStrength(currentCorrComm, otherCommunity);
				
				// connCorrComm==connSmaller || connCorrComm==connBigger ||  
				if (connCorrComm==connSmaller+connBigger)
					myAffected.add(currentCorrComm);
			}
			other.removeAll(otherAffected.get(otherDimension));
			
			
			otherConnected.put(otherDimension, other);
			
			
			
			Set<Integer> my = otherEdgesToMe.getConnectedFromAny(otherAffected.get(otherDimension));
			
			// brauchen wir das hier? definitiv notwendig bislang nur die andere Richtung
			/*
			Map<Integer, CommunityWithStrength> corrCommMy = 
				((MurataDomainState)this.domainStates.get(dimension)).correspondingCommunities.get(otherDimension);		
			
			Set<Integer> otherAff = otherAffected.get(otherDimension);			
			for (int myCommunity: my)
			{
				otherAff.add(corrCommMy.get(myCommunity).community);
			}
			*/
			
			my.removeAll(myAffected);
			myConnected.put(otherDimension, my);

		}

		
		/** second step: update modularities and corresponding communities */ 
		// optimization: we're catching all affected communities right in computeGain, makes things easier here
		performUpdates(joinPair);
		
		
		/** third step: join edges */
		for (DomainState domainstate: this.domainStates.values())
			domainstate.join(joinPair.dimension, joinPair.pair);
		
		
		/** fourth step: finally update gain */
		gtm.get(dimension).removeElement(biggerElement);
		myAffected.remove(biggerElement);
		
		// initialize excessive bookkeeping
		this.pairsComputed = new int[this.numDimensions];
		this.pairsChanged = new int[this.numDimensions];
		this.pairsMax = new int[this.numDimensions];
		this.connectedPairsComputed = new int[this.numDimensions];
		this.connectedPairsChanged = new int[this.numDimensions];
		this.affectedPairsComputed = new int[this.numDimensions];
		this.affectedPairsChanged = new int[this.numDimensions];
		
		// update my dimension
		Set<Integer> myAll = this.assignments.get(dimension).byCommunity().keySet();
		GainTracker gt = gtm.get(dimension);
		updateGainAffectedPairs(dimension, myAll, myAffected, gt);
		updateGainConnectedPairList(dimension, myConnected, gt);
		
		// update other dimension
		for (Map.Entry<Integer, Set<Integer>> e: otherAffected.entrySet())
		{
			int otherDimension = e.getKey();
			Set<Integer> affected = e.getValue();
			Set<Integer> otherAll = this.assignments.get(otherDimension).byCommunity().keySet(); 
			gt = gtm.get(otherDimension);
			updateGainAffectedPairs(otherDimension, otherAll, affected, gt);
			updateGainConnectedPairs(otherDimension, otherConnected.get(otherDimension), gt);
		}
	}
	
	// ok so connectedness is getting slightly complicated for my dimension
	// because we have to sources of connectedness pairs, the two (or n-1) dimensions
	// imagine due to dim 1, nodes 1,2,3 are connected, due to dime 2, nods 1,2,4 are connected
	// in this case, 1,2 would be recomputed twice if we simply called updateGainConnectedPairs
	// so we create a map where each to-be-computed pair is stored and if it gets entered more
	// than once, we don't care
	// could also use a set but not sure about identity properties and it's surely not faster than this special approach	
	void updateGainConnectedPairList(int dimension, Map<Integer, Set<Integer>> connectedMap, GainTracker gt)
	{
		// create a list of all pairs
		Map<Integer, Set<Integer>> pairs = new HashMap<Integer, Set<Integer>>();
		for (Set<Integer> connected: connectedMap.values())
		{
			Integer[] connectedList = connected.toArray(new Integer[0]);
			for (int i=0; i<connectedList.length; i++)
			{
				int com1 = connectedList[i];
				for (int j=i+1; j<connectedList.length; j++)
				{
					int com2 = connectedList[j];			
					
					int smaller = Math.min(com1, com2);
					int bigger = Math.max(com1, com2);
					
					Set<Integer> biggers = null;
					if (pairs.containsKey(smaller))
					{
						biggers = pairs.get(smaller);
					}
					else
					{
						biggers = new HashSet<Integer>();
						pairs.put(smaller, biggers);
					}
					biggers.add(bigger);
				}
			}
		}
		
		// compute updates
		for (Map.Entry<Integer, Set<Integer>> e: pairs.entrySet())
		{
			int smaller = e.getKey();
			for (int bigger: e.getValue())
			{
				Pair pair = new Pair(smaller, bigger);
				JoinPair jp = new JoinPair(pair, dimension);
				double newGain = computeGain(jp, false);
				
				this.pairsComputed[dimension] += 1;
				this.connectedPairsComputed[dimension] += 1;
				
				if (gt.getPairGain(pair)!=newGain)
				{
					this.pairsChanged[dimension] += 1;
					this.connectedPairsChanged[dimension] += 1;
					gt.setPair(pair, newGain);
				}									
			}
		}
	}
	
	
	void updateGainConnectedPairs(int dimension, Set<Integer> connected, GainTracker gt)
	{
		Integer[] connectedList = connected.toArray(new Integer[0]);
		for (int i=0; i<connectedList.length; i++)
		{
			int com1 = connectedList[i];
			for (int j=i+1; j<connectedList.length; j++)
			{
				int com2 = connectedList[j];

				Pair pair = new Pair(Math.min(com1, com2),
						Math.max(com1, com2));
				JoinPair jp = new JoinPair(pair, dimension);
				double newGain = computeGain(jp, false);
				
				this.pairsComputed[dimension] += 1;
				this.connectedPairsComputed[dimension] += 1;
				
				if (gt.getPairGain(pair)!=newGain)
				{
					this.pairsChanged[dimension] += 1;
					this.connectedPairsChanged[dimension] += 1;
					gt.setPair(pair, newGain);
				}					
			}
		}		
	}
	
	
	void updateGainAffectedPairs(int dimension, Set<Integer> all, Set<Integer> affected, GainTracker gt)
	{
		int m = all.size();
		m = (m*(m-1))/2;
		this.pairsMax[dimension] = m;
		for (int otherCommunity: all)
		{
			Boolean otherIsAffected = affected.contains(otherCommunity);
			for (int affectedCommunity: affected)
			{
				if (otherCommunity == affectedCommunity)
					continue;
				if (otherIsAffected && affectedCommunity<otherCommunity)
					continue;
				Pair pair = new Pair(Math.min(otherCommunity, affectedCommunity),
						Math.max(otherCommunity, affectedCommunity));
				JoinPair jp = new JoinPair(pair, dimension);
				double newGain = computeGain(jp, false);
				
				this.pairsComputed[dimension] += 1;
				this.affectedPairsComputed[dimension] += 1;
				
				if (gt.getPairGain(pair)!=newGain)
				{
					this.pairsChanged[dimension] += 1;
					this.affectedPairsChanged[dimension] += 1;
					gt.setPair(pair, newGain);
				}
			}
		}
	}
	
	
}
